from __future__ import unicode_literals

from decimal import Decimal
from dateutil.parser import parse

from fakturoid import six

__all__ = ['Account', 'Subject', 'InvoiceLine', 'Invoice', 'Generator', 'Message']


class Model(six.UnicodeMixin):
    """Base class for all Fakturoid model objects"""
    id = None

    def __init__(self, **fields):
        self.update(fields)

    def __repr__(self):
        return "<{0}:{1}>".format(self.__class__.__name__, self.id)

    def update(self, fields):
        for field, value in fields.items():
            if value and isinstance(value, six.string_types):
                if field.endswith('_at'):
                    fields[field] = parse(value)
                elif field.endswith('_on') or field.endswith('_due') or field.endswith('_date'):
                    fields[field] = parse(value).date()
                elif field in self.Meta.decimal:
                    fields[field] = Decimal(value)
        self.__dict__.update(fields)

    def is_field_writable(self, field, value):
        # if hasattr(self.Meta, 'writable'):
        #    return field in self.Meta.writable
        if hasattr(self.Meta, 'readonly'):
            return field not in self.Meta.readonly
        return True

    def serialize_field_value(self, field, value):
        if isinstance(value, Model):
            return value.get_fields()
        if isinstance(value, list):
            nv = []
            for i, item in enumerate(value):
                nv.append(self.serialize_field_value('{0}.{1}'.format(field, i), item))
            return nv
        if isinstance(value, Decimal):
            return str(value)
        if hasattr(value, 'isoformat'):
            return value.isoformat()
        return value

    def get_fields(self):
        data = {}
        for field, value in self.__dict__.items():
            if self.is_field_writable(field, value):
                data[field] = self.serialize_field_value(field, value)
        return data


class Account(Model):
    """See http://docs.fakturoid.apiary.io/ for complete field reference."""
    name = None

    class Meta:
        decimal = []

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return "<{0}:{1}>".format(self.__class__.__name__, self.name)


class Subject(Model):
    """See http://docs.fakturoid.apiary.io/ for complete field reference."""
    name = None

    class Meta:
        readonly = ['id', 'avatar_url', 'html_url', 'url', 'updated_at']
        decimal = []

    def __unicode__(self):
        return self.name


class InvoiceLine(Model):
    quantity = None

    class Meta:
        readonly = []  # no id here, to correct update
        decimal = ['quantity', 'unit_price']

    def __init__(self, **kwargs):
        self.quantity = Decimal(1)
        super(InvoiceLine, self).__init__(**kwargs)

    def __unicode__(self):
        if self.unit_name:
            return "{0} {1} {2}".format(self.quantity, self.unit_name, self.name)
        else:
            if self.quantity == 1:
                return self.name
            else:
                return "{0} {1}".format(self.quantity, self.name)


class AbstractInvoice(Model):
    lines = []
    _loaded_lines = []  # keep loaded data to be able delete removed lines

    def update(self, fields):
        if 'lines' in fields:
            self.lines = []
            self._loaded_lines = []
            for line in fields.pop('lines'):
                if not isinstance(line, InvoiceLine):
                    if 'id' in line:
                        self._loaded_lines.append(line)
                    line = InvoiceLine(**line)
                self.lines.append(line)
        super(AbstractInvoice, self).update(fields)

    def serialize_field_value(self, field, value):
        result = super(AbstractInvoice, self).serialize_field_value(field, value)
        if field == 'lines':
            ids = map(lambda l: l.get('id'), result)
            for remote in self._loaded_lines:
                if remote['id'] not in ids:
                    remote['_destroy'] = True
                    result.append(remote)
        return result

    def is_field_writable(self, field, value):
        if field.startswith('your_') or field.startswith('client_'):
            return False
        return super(AbstractInvoice, self).is_field_writable(field, value)


class Invoice(AbstractInvoice):
    """See http://docs.fakturoid.apiary.io/ for complete field reference."""
    number = None

    class Meta:
        readonly = [
            'id', 'token', 'status', 'due_on',
            'sent_at', 'paid_at', 'reminder_sent_at', 'accepted_at', 'canceled_at',
            'subtotal', 'native_subtotal', 'total', 'native_total',
            'remaining_amount', 'remaining_native_amount',
            'html_url', 'public_html_url', 'url', 'updated_at',
            'subject_url'
        ]
        decimal = [
            'exchange_rate', 'subtotal', 'total',
            'native_subtotal', 'native_total', 'remaining_amount',
            'remaining_native_amount'
        ]

    def __unicode__(self):
        return self.number


class Generator(AbstractInvoice):
    """See http://docs.fakturoid.apiary.io/ for complete field reference."""
    name = None

    class Meta:
        readonly = [
            'id', 'subtotal', 'native_subtotal', 'total', 'native_total',
            'html_url', 'url', 'subject_url', 'updated_at'
        ]
        decimal = ['exchange_rate', 'subtotal', 'total', 'native_subtotal', 'native_total']

    def __unicode__(self):
        return self.name


class Message(Model):
    """See http://docs.fakturoid.apiary.io/#reference/messages for complete field reference."""
    subject = None

    class Meta:
        decimal = []

    def __unicode__(self):
        return self.subject



import sys

if sys.version_info[0] >= 3:  # Python 3
    string_types = (str,)
else:
    string_types = (basestring,)


class UnicodeMixin(object):
    """Mixin class to handle defining the proper __str__/__unicode__
    methods in Python 2 or 3."""
    if sys.version_info[0] >= 3:  # Python 3
        def __str__(self):
            return self.__unicode__()
    else:  # Python 2
        def __str__(self):
            return self.__unicode__().encode('utf8')



from itertools import islice

from fakturoid import six


class PagedResource(object):
    """List adapter for paged resources. Returns sliceable lazy loaded object."""

    def __init__(self, page_size=20):
        self.pages = {}
        self.page_size = page_size or 20
        self.page_count = None

    def load_page(self, n):
        raise NotImplementedError("You must implement load_page method.")

    def ensure_page_count(self):
        if self.page_count is None:
            # load any page to get page count from headers
            self.get_page(0)

    def get_page(self, n):
        if self.page_count and n >= self.page_count:
            raise IndexError('index out of range')
        if n in self.pages:
            return self.pages[n]
        page = self.load_page(n)
        if page:
            self.pages[n] = page
            return page
        raise IndexError('index out of range')

    def __len__(self):
        self.ensure_page_count()
        return (self.page_size * (self.page_count - 1) +
                len(self.get_page(self.page_count - 1)))

    def __getitem__(self, key):
        if isinstance(key, int):
            if key < 0:
                key = len(self) + key
                if key < 0:
                    raise IndexError('index out of range')
            page_n, idx = divmod(key, self.page_size)
            return self.get_page(page_n)[idx]
        elif isinstance(key, slice):
            # TODO support negative step
            return islice(self, *key.indices(len(self)))
        else:
            raise TypeError('list indices must be integers')


class ModelList(PagedResource, six.UnicodeMixin):

    def __init__(self, model_api, endpoint, params=None):
        super(ModelList, self).__init__()
        self.model_api = model_api
        self.endpoint = endpoint
        self.params = params or {}

    def load_page(self, n):
        params = {'page': n + 1}
        params.update(self.params)
        response = self.model_api.session._get(self.endpoint, params=params)
        if self.page_count is None:
            self.page_count = response.get('page_count', n + 1)
        return list(self.model_api.unpack(response))

    def __unicode__(self):
        # TODO print if loaded
        return "<list of {0} models>".format(self.model_api.model_type.__name__)



import os
from setuptools import setup

# work around to prevent http://bugs.python.org/issue15881 from showing up
try:
    import multiprocessing
except ImportError:
    pass

# def read(fname):
#     return open(os.path.join(os.path.dirname(__file__), fname)).read()

long_description = """
fakturoid.cz Python API
=======================

The Python interface to online accounting service
`Fakturoid <http://fakturoid.cz/>`_.

See documentation on https://github.com/farin/python-fakturoid
"""

setup(
    name='fakturoid',
    version='1.1',
    url="https://github.com/farin/python-fakturoid",
    description='Python API for fakturoid.cz',
    # long_description=read('README.md'),
    long_description=long_description,
    author='Roman Krejcik',
    author_email='farin@farin.cz',
    license='MIT',
    platforms='any',
    keywords=['fakturoid', 'accounting'],
    packages=['fakturoid'],
    install_requires=['requests', 'python-dateutil'],
    tests_require=['mock'],
    test_suite="tests",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Financial :: Accounting",
    ],
)



import re
import json
from datetime import date, datetime
from functools import wraps

import requests

from fakturoid.models import Account, Subject, Invoice, Generator, Message
from fakturoid.paging import ModelList

__all__ = ['Fakturoid']

link_header_pattern = re.compile(r'page=(\d+)[^>]*>; rel="last"')


class Fakturoid(object):
    """Fakturoid API v2 - http://docs.fakturoid.apiary.io/"""
    slug = None
    api_key = None
    user_agent = 'python-fakturoid (https://github.com/farin/python-fakturoid)'

    _models_api = None

    def __init__(self, slug, email, api_key, user_agent=None):
        self.slug = slug
        self.api_key = api_key
        self.email = email
        self.user_agent = user_agent or self.user_agent

        self._models_api = {
            Account: AccountApi(self),
            Subject: SubjectsApi(self),
            Invoice: InvoicesApi(self),
            Generator: GeneratorsApi(self),
            Message: MessagesApi(self),
        }

    def model_api(model_type=None):
        def wrap(fn):
            @wraps(fn)
            def wrapper(self, *args, **kwargs):
                mt = model_type or type(args[0])
                mapi = self._models_api.get(mt)
                if not mapi:
                    raise TypeError('model expected, got {0}'.format(mt.__name__))
                return fn(self, mapi, *args, **kwargs)
            return wrapper
        return wrap

    def account(self):
        return self._models_api[Account].load()

    @model_api(Subject)
    def subject(self, mapi, id):
        return mapi.load(id)

    @model_api(Subject)
    def subjects(self, mapi, *args, **kwargs):
        return mapi.find(*args, **kwargs)

    @model_api(Invoice)
    def invoice(self, mapi, id):
        return mapi.load(id)

    @model_api(Invoice)
    def invoices(self, mapi, *args, **kwargs):
        return mapi.find(*args, **kwargs)

    @model_api(Generator)
    def generator(self, mapi, id):
        return mapi.load(id)

    @model_api(Generator)
    def generators(self, mapi, *args, **kwargs):
        return mapi.find(*args, **kwargs)

    @model_api()
    def save(self, mapi, obj, **kwargs):
        mapi.save(obj, **kwargs)

    @model_api()
    def delete(self, mapi, obj):
        """Call with loaded model or use new instance directly.
        s = fa.subject(1234)
        a.delete(s)

        fa.delete(Subject(id=1234))
        """
        mapi.delete(obj)

    def _extract_page_link(self, header):
        m = link_header_pattern.search(header)
        if m:
            return int(m.group(1))
        return None

    def _make_request(self, method, success_status, endpoint, **kwargs):
        url = "https://app.fakturoid.cz/api/v2/accounts/{0}/{1}.json".format(self.slug, endpoint)
        headers = {'User-Agent': self.user_agent}
        headers.update(kwargs.pop('headers', {}))
        r = getattr(requests, method)(url, auth=(self.email, self.api_key), headers=headers, **kwargs)
        try:
            json_result = r.json()
        except:
            json_result = None

        if r.status_code == success_status:
            response = {'json': json_result}
            if 'link' in r.headers:
                page_count = self._extract_page_link(r.headers['link'])
                if page_count:
                    response['page_count'] = page_count
            return response

        if json_result and "errors" in json_result:
            raise ValueError(json_result["errors"])

        r.raise_for_status()

    def _get(self, endpoint, params=None):
        return self._make_request('get', 200, endpoint, params=params)

    def _post(self, endpoint, data):
        return self._make_request('post', 201, endpoint, headers={'Content-Type': 'application/json'}, data=json.dumps(data))

    def _put(self, endpoint, data):
        return self._make_request('put', 200, endpoint, headers={'Content-Type': 'application/json'}, data=json.dumps(data))

    def _delete(self, endpoint):
        return self._make_request('delete', 204, endpoint)


class ModelApi(object):
    session = None
    model_type = None
    endpoint = None

    def __init__(self, session):
        self.session = session

    def extract_id(self, value):
        if isinstance(value, int):
            return value
        if not isinstance(value, self.model_type):
            raise TypeError("int or {0} expected".format(self.model_type.__name__.lower()))
        if not getattr(value, 'id', None):
            raise ValueError("object wit unassigned id")
        return value.id

    def unpack(self, response):
        raw = response['json']
        if isinstance(raw, list):
            objects = []
            for fields in raw:
                objects.append(self.model_type(**fields))
            return objects
        else:
            return self.model_type(**raw)


class CrudModelApi(ModelApi):
    def load(self, id):
        if not isinstance(id, int):
            raise TypeError('id must be int')
        response = self.session._get('{0}/{1}'.format(self.endpoint, id))
        return self.unpack(response)

    def find(self, params={}, endpoint=None):
        response = self.session._get(endpoint or self.endpoint, params=params)
        return self.unpack(response)

    def save(self, model):
        if model.id:
            result = self.session._put('{0}/{1}'.format(self.endpoint, model.id), model.get_fields())
        else:
            result = self.session._post(self.endpoint, model.get_fields())
        model.update(result['json'])

    def delete(self, model):
        id = self.extract_id(model)
        self.session._delete('{0}/{1}'.format(self.endpoint, id))


class AccountApi(ModelApi):
    model_type = Account
    endpoint = 'account'

    def load(self):
        response = self.session._get(self.endpoint)
        return self.unpack(response)


class SubjectsApi(CrudModelApi):
    model_type = Subject
    endpoint = 'subjects'

    def find(self, since=None):
        params = {}
        if since:
            if not isinstance(since, (datetime, date)):
                raise TypeError("'since' parameter must be date or datetime")
            params['since'] = since.isoformat()
        return super(SubjectsApi, self).find(params)


class InvoicesApi(CrudModelApi):
    """If number argument is givent returms single Invoice object (or None), otherwise iterable list of invoices are returned.
    """
    model_type = Invoice
    endpoint = 'invoices'

    STATUSES = ['open', 'sent', 'overdue', 'paid', 'cancelled']

    def find(self, proforma=None, subject_id=None, since=None, updated_since=None, number=None, status=None):
        params = {}
        if subject_id:
            if not isinstance(subject_id, int):
                raise TypeError("'subject_id' parameter must be int")
            params['subject_id'] = subject_id
        if since:
            if not isinstance(since, (datetime, date)):
                raise TypeError("'since' parameter must be date or datetime")
            params['since'] = since.isoformat()
        if updated_since:
            if not isinstance(updated_since, (datetime, date)):
                raise TypeError("'updated_since' parameter must be date or datetime")
            params['updated_since'] = updated_since.isoformat()
        if number:
            params['number'] = number
        if status:
            if status not in self.STATUSES:
                raise ValueError('invalid invoice status, expected one of {0}'.format(', '.join(self.STATUSES)))
            params['status'] = status

        if proforma is None:
            endpoint = self.endpoint
        elif proforma:
            endpoint = '{0}/proforma'.format(self.endpoint)
        else:
            endpoint = '{0}/regular'.format(self.endpoint)

        return ModelList(self, endpoint, params)


class GeneratorsApi(CrudModelApi):
    model_type = Generator
    endpoint = 'generators'

    def find(self, recurring=None, subject_id=None, since=None):
        params = {}
        if subject_id:
            if not isinstance(subject_id, int):
                raise TypeError("'subject_id' parameter must be int")
            params['subject_id'] = subject_id
        if since:
            if not isinstance(since, (datetime, date)):
                raise TypeError("'since' parameter must be date or datetime")
            params['since'] = since.isoformat()

        if recurring is None:
            endpoint = self.endpoint
        elif recurring:
            endpoint = '{0}/recurring'.format(self.endpoint)
        else:
            endpoint = '{0}/template'.format(self.endpoint)

        return super(GeneratorsApi, self).find(params, endpoint)


class MessagesApi(ModelApi):
    model_type = Message
    endpoint = 'message'

    def save(self, model, **kwargs):
        invoice_id = kwargs.get('invoice_id')
        if not isinstance(invoice_id, int):
            raise TypeError("invoice_id must be int")
        result = self.session._post('invoices/{0}/{1}'.format(invoice_id, self.endpoint), model.get_fields())
        model.update(result['json'])
